------------------------------------------------------------------------
--  PROCESSADOR R6
--
--  Fernando Moraes - versao inicial - 01/10/2000
--                  - ultima alteracao - 15/10/2000
--            
------------------------------------------------------------------------

--------------------------------------------------------------------------
-- package com tipos basicos, e funcao de soma 
-------------------------------------------------------------------------
library IEEE;
use IEEE.Std_Logic_1164.all;

package r6 is              
	subtype reg16  is std_logic_vector(15 downto 0);       
	subtype reg8   is std_logic_vector(7 downto 0);
	subtype reg4   is std_logic_vector(3 downto 0);       
	
	type microinstrucao is record                 
		wir:      std_logic;      -- controle de escrita nos registradores
		rs2:      std_logic; 
		wreg:     std_logic; 
		wpc:      std_logic; 
		wsp:      std_logic;      -- controle de escrita na PILHA
		incsp:    std_logic;      
		orgsp:    std_logic;      
		destsp:   std_logic;
		sela:     std_logic;		-- selecao de quem VAI para a ULA
		selb:     std_logic;      
		din:      std_logic;      -- selecao de qem VEM  para a memoria
		ad:       std_logic;      -- selecao do ENDERECO para a memoria 
		mula:     std_logic;	    -- selecao de quem ESCREVE na memoria  
		mopB:     std_logic;
		ce,rw:    std_logic;      -- Chip enable e R_Wnegado       
		ula:      reg4;           -- operacao realizada na ULA
	end record;
	
	constant d0  : reg4 := "0000";       -- digito 0, digito 1, digito 2, etc.
	constant d1  : reg4 := "0001";
	constant d2  : reg4 := "0010";
	constant d3  : reg4 := "0011";       
	constant d4  : reg4 := "0100";
	constant d5  : reg4 := "0101";
	constant d6  : reg4 := "0110";
	constant d7  : reg4 := "0111";
	constant d8  : reg4 := "1000";
	constant d9  : reg4 := "1001";
	constant d10 : reg4 := "1010";       
	constant d11 : reg4 := "1011";
	constant d12 : reg4 := "1100";
	constant d13 : reg4 := "1101";
	constant d14 : reg4 := "1110";
	constant d15 : reg4 := "1111";       
	
	constant incB  : reg4 := "0000";     
	constant add   : reg4 := "0001";
	constant sub   : reg4 := "0010";
	constant ou    : reg4 := "0011";       
	constant e     : reg4 := "0100";
	constant ouX   : reg4 := "0101";
	constant addi  : reg4 := "0110";
	constant subi  : reg4 := "0111";
	constant cte_L : reg4 := "1000";
	constant cte_H : reg4 := "1001";      
	constant sl    : reg4 := "1011";
	constant sr    : reg4 := "1100";
	constant notB  : reg4 := "1101";
	constant jmp   : reg4 := "1110";
	constant passaA: reg4 := "1111";
	
	component reg16clear 
		port( ck, rst, ce: in std_logic; D: in reg16; Q: out reg16 );
	end component;
end r6;

--------------------------------------------------------------------------
-- Registrador de uso geral         -   SENSIVEL A BORDA DE DESCIDA
--------------------------------------------------------------------------
library IEEE;
use IEEE.Std_Logic_1164.all;
use work.r6.all;

entity reg16clear is
	port( ck,rst,ce:in std_logic;   D:in reg16;  Q:out reg16 );
end reg16clear;

architecture reg16clear of reg16clear is
	begin
	process (ck, rst)
		begin
		if rst = '1' then
			Q <= (others => '0');
		elsif ck'event and ck = '0' then  
			if ce = '1' then  Q <= D; end if;
		end if;
	end process;
end reg16clear;

--------------------------------------------------------------------------
-- Registrador SP
--------------------------------------------------------------------------
library IEEE;
use IEEE.Std_Logic_1164.all;
use IEEE.Std_Logic_unsigned.all;
use work.r6.all;

entity reg16sp is
	port( ck, rst, wSP, incSP, orgSP, destSP: in std_logic; 
		inSP: in reg16; 
		outSP: out reg16 );
end reg16sp;

architecture reg16sp of reg16sp is
	signal cout, coutant, c0: std_logic;
	signal ssp, s1, s2: reg16;    --Saida do SP (ver folha)
	begin           
	s1 <= ssp+1 when incsp = '1' else ssp-1;         -- auto-incremento ou decremento    
	s2 <=  inSP when orgSP = '0' else s1;           -- selecao da origem do dado     
	r0:  reg16clear port map(ck=>ck, rst=>rst, ce=>wSP ,d=>s2,q=>ssp );                                                                                                       
	outSP <= ssp when destSP = '0' else s1;          -- selecao da saida do dado   
end reg16sp;

--------------------------------------------------------------------------
-- Banco de registradores (R0 .. R15)
--------------------------------------------------------------------------
library IEEE;
use IEEE.Std_Logic_1164.all;
use work.r6.all;
entity bcregs is
	port(  ck, rst, wreg, rs2: in std_logic;       
		ir, inREG:          in reg16;
		source1, source2:   out reg16
		);
end entity;

architecture bcregs of bcregs is     
	signal reg1, reg2,  reg3,  reg4a, reg5,  reg6,  reg7,  reg8, 
	reg9, reg10, reg11, reg12, reg13, reg14, reg15: reg16;        
	signal wen: std_logic_vector(15 downto 1); 
	signal destB: reg4; 
	begin    
	
	r1:  reg16clear port map(ck=>ck,rst=>rst,ce=>wen(1) ,d=>inREG,q=>reg1 );   
	r2:  reg16clear port map(ck=>ck,rst=>rst,ce=>wen(2) ,d=>inREG,q=>reg2 );
	r3:  reg16clear port map(ck=>ck,rst=>rst,ce=>wen(3) ,d=>inREG,q=>reg3 );
	r4:  reg16clear port map(ck=>ck,rst=>rst,ce=>wen(4) ,d=>inREG,q=>reg4a);  
	r5:  reg16clear port map(ck=>ck,rst=>rst,ce=>wen(5) ,d=>inREG,q=>reg5 );    
	r6:  reg16clear port map(ck=>ck,rst=>rst,ce=>wen(6) ,d=>inREG,q=>reg6 );
	r7:  reg16clear port map(ck=>ck,rst=>rst,ce=>wen(7) ,d=>inREG,q=>reg7 );
	r8:  reg16clear port map(ck=>ck,rst=>rst,ce=>wen(8) ,d=>inREG,q=>reg8 );  
	r9:  reg16clear port map(ck=>ck,rst=>rst,ce=>wen(9) ,d=>inREG,q=>reg9 );    
	r10: reg16clear port map(ck=>ck,rst=>rst,ce=>wen(10),d=>inREG,q=>reg10);
	r11: reg16clear port map(ck=>ck,rst=>rst,ce=>wen(11),d=>inREG,q=>reg11);
	r12: reg16clear port map(ck=>ck,rst=>rst,ce=>wen(12),d=>inREG,q=>reg12);  
	r13: reg16clear port map(ck=>ck,rst=>rst,ce=>wen(13),d=>inREG,q=>reg13);    
	r14: reg16clear port map(ck=>ck,rst=>rst,ce=>wen(14),d=>inREG,q=>reg14);
	r15: reg16clear port map(ck=>ck,rst=>rst,ce=>wen(15),d=>inREG,q=>reg15);
	
	-- escolha de qual registrador vai ser alterado - NAO ESCREVE NO REG0! ---    
	wen(1)  <=  '1' when  ir(11 downto 8)=d1  and wreg='1'  else '0';      
	wen(2)  <=  '1' when  ir(11 downto 8)=d2  and wreg='1'  else '0';      
	wen(3)  <=  '1' when  ir(11 downto 8)=d3  and wreg='1'  else '0';      
	wen(4)  <=  '1' when  ir(11 downto 8)=d4  and wreg='1'  else '0';      
	wen(5)  <=  '1' when  ir(11 downto 8)=d5  and wreg='1'  else '0';      
	wen(6)  <=  '1' when  ir(11 downto 8)=d6  and wreg='1'  else '0';      
	wen(7)  <=  '1' when  ir(11 downto 8)=d7  and wreg='1'  else '0';      
	wen(8)  <=  '1' when  ir(11 downto 8)=d8  and wreg='1'  else '0';      
	wen(9)  <=  '1' when  ir(11 downto 8)=d9  and wreg='1'  else '0';      
	wen(10) <=  '1' when  ir(11 downto 8)=d10 and wreg='1'  else '0';      
	wen(11) <=  '1' when  ir(11 downto 8)=d11 and wreg='1'  else '0';      
	wen(12) <=  '1' when  ir(11 downto 8)=d12 and wreg='1'  else '0';      
	wen(13) <=  '1' when  ir(11 downto 8)=d13 and wreg='1'  else '0';      
	wen(14) <=  '1' when  ir(11 downto 8)=d14 and wreg='1'  else '0';       
	wen(15) <=  '1' when  ir(11 downto 8)=d15 and wreg='1'  else '0'; 	 
	
	-- selecao do source1
	source1 <=  (others=> '0')  when  ir(7 downto 4)=d0  else
	reg1  when  ir(7 downto 4)=d1  else
	reg2  when  ir(7 downto 4)=d2  else
	reg3  when  ir(7 downto 4)=d3  else
	reg4a when  ir(7 downto 4)=d4  else
	reg5  when  ir(7 downto 4)=d5  else
	reg6  when  ir(7 downto 4)=d6  else
	reg7  when  ir(7 downto 4)=d7  else
	reg8  when  ir(7 downto 4)=d8  else
	reg9  when  ir(7 downto 4)=d9  else
	reg10 when  ir(7 downto 4)=d10 else
	reg11 when  ir(7 downto 4)=d11 else
	reg12 when  ir(7 downto 4)=d12 else
	reg13 when  ir(7 downto 4)=d13 else
	reg14 when  ir(7 downto 4)=d14 else
	reg15;    
	
	-- decodificacao para o source2	  
	destB <= ir(3 downto 0) when rs2 = '0' else ir(11 downto 8);
	source2 <=  (others=> '0')  when destB=d0  else
	reg1  when destB=d1  else
	reg2  when destB=d2  else
	reg3  when destB=d3  else
	reg4a when destB=d4  else
	reg5  when destB=d5  else
	reg6  when destB=d6  else
	reg7  when destB=d7  else
	reg8  when destB=d8  else
	reg9  when destB=d9  else
	reg10 when destB=d10 else
	reg11 when destB=d11 else
	reg12 when destB=d12 else
	reg13 when destB=d13 else
	reg14 when destB=d14 else
	reg15;
end bcregs;

--------------------------------------------------------------------------
--------------------------------------------------------------------------
--  Descricao estrutural do bloco de dados
--------------------------------------------------------------------------
--------------------------------------------------------------------------
library IEEE;
use IEEE.Std_Logic_1164.all;
use IEEE.Std_Logic_unsigned.all;
use work.r6.all;

entity datapath is
	port( uins: in microinstrucao;
		ck, rst: in std_logic;
		instrucao, endereco : out reg16;
		data: inout reg16;
		flag: out std_logic
		); 
end datapath;

architecture datapath of datapath is
	
	component bcregs
		port( ck, rst, wreg, rs2: in std_logic;       
			ir, inREG:           in reg16;
			source1, source2:    out reg16
			); 
	end component;         
	
	component reg16sp
		port( ck, rst, wSP, incSP, orgSP, destSP: in std_logic; 
			inSP: in reg16; 
			outSP: out reg16 );
	end component;
	
	signal extended4, extended8, out_ula, s1, s2, mux1, ir, pc, sp, opA, opB: reg16;
	signal igual,dif,maior,menor,iflag: std_logic;
	begin  
	
	REGS: bcregs port map(ck=>ck,rst=>rst, wreg=>uins.wreg, rs2=>uins.rs2, 
		ir=>ir, inREG=> mux1, 
		source1=>s1, source2=>s2);
	
	RPC: reg16clear port map(ck=>ck,rst=>rst,ce=>uins.wpc ,d=>mux1,q=>pc ); 
	
	RSP: reg16sp port map(ck=>ck,rst=>rst, wSP=>uins.wsp, incSP=>uins.incsp, orgSP=>uins.orgsp, destSP=>uins.destSP,
		inSP=>out_ula,
		outSP=>sp);
	
	-- registrador IR
	RIR:  reg16clear port map(ck=>ck,rst=>rst,ce=>uins.wir ,d=>data,q=>ir ); 
	instrucao <= ir;
	
	-- operando A para a ULA
	opA <= s1 when uins.selA = '1' else ir;
	
	-- operando B para a ULA
	opB <= s2 when uins.selB = '1' else pc;
	
	-- ENDERECO       para a memoria	(mux2)
	endereco <= opB when uins.ad = '0' else sp;
	
	-- conteudo que vai para os registradores ou PC (mux1)
	mux1 <=  out_ula when uins.din= '1' else data;         
	
	-- DADOS que serao escritos na memoria
	data <= out_ula when uins.mula='1'  else (others => 'Z');
	data <= opB     when uins.mopb='1'  else (others => 'Z');
	
	--
	--  extensao de sinal, necessario para imediatos e saltos
	--
	extended8 <= opA(7) & opA(7) & opA(7) & opA(7) & opA(7) & opA(7) & opA(7) & opA(7) & opA(7 downto 0);
	extended4 <= opA(11) & opA(11) & opA(11) & opA(11) & opA(11 downto 0);  
	
	---
	---  ULA
	---                
	out_ula <= opB + '1'                          when uins.ula=incB   else 
	opA + opB                          when uins.ula=add    else	 
	opA - opB                          when uins.ula=sub    else
	opA or opB                         when uins.ula=ou     else
	opA and opB                        when uins.ula=e      else  
	opA xor opB                        when uins.ula=ouX    else
	opB + extended8                    when uins.ula=addi   else	 
	opB - extended8                    when uins.ula=subi   else
	opB(15 downto 8) & opA(7 downto 0) when uins.ula=cte_L  else
	opA(7 downto 0)  & opB(7 downto 0) when uins.ula=cte_H  else 
	
	opB(14 downto 0) & iflag           when uins.ula=sl     else
	iflag & opB(15 downto 1)           when uins.ula=sr     else	 
	not opB(15 downto 0)               when uins.ula=notB   else				  
	
	opB + extended4                   when uins.ula=jmp    else
	opA;            -- DEFAULT: passaA	  	             
	
	---
	--- armazenamento do flag -- ATENCAO: TAMBEM NA BORDA DE DESCIDA
	---	  
	flag <= iflag;	
	igual <=  '1' when opA =  opB else '0';
	dif   <=  '1' when opA /= opB else '0';
	maior <=  '1' when opA >  opB else '0';
	menor <=  '1' when opA <  opB else '0';
	process (ck, rst)
		begin
		if rst = '1' then
			iflag <= '0';
		elsif ck'event and ck = '0' then  
			if ir(15 downto 8)=x"D0" then     iflag <= dif;
			elsif ir(15 downto 8)=x"D1" then	 iflag <= igual;
			elsif ir(15 downto 8)=x"D2" then	 iflag <= maior;
			elsif ir(15 downto 8)=x"D3" then	 iflag <= menor;
			end if;
		end if;
	end process;                          
	
end datapath;

--------------------------------------------------------------------------
--------------------------------------------------------------------------
--  Descricao comportamental do bloco de controle - FORMATADA CORRETAMENTE
--------------------------------------------------------------------------
--------------------------------------------------------------------------
library IEEE;
use IEEE.Std_Logic_1164.all;
use work.r6.all;
entity controle is
	port (uins:    out microinstrucao;
		rst,ck: in std_logic;
		flag:   in std_logic;
		ir:     in reg16
		);
end controle;

architecture controle of controle is   
	type type_state  is (FETCH, EXECUTA, HALT);
    signal estado :  type_state;
	signal c0, c1: std_logic;
	signal op1, op2, op3, op4: reg4;
	begin
	
	-- os ops, c1 e c0 sao colocados apenas como auxilio para facilita	a decodificacao
	op1 <= ir(15 downto 12);
	op2 <= ir(11 downto  8);
	op3 <= ir( 7 downto  4);
	op4 <= ir( 3 downto  0);          
	c0 <= '0';
	c1 <= '1';    
	
    --
	-- DEFINE O SEQUENCIAMENTO DAS MICROINSTRUCOES 
	--
	process(rst, ck)
		begin
		
		if rst='1' then
			estado <= EXECUTA;  -- primeiro borda de subida de clock depois do reset
			                    -- fara um fetch
			
		elsif  ck'event and ck='1' then	   
			
			case estado is
				  when FETCH =>   if ir=x"D700" then	-- HALT => nao gera mais microinstrucoes
					                estado <= HALT;
				                   else
					                estado <= EXECUTA;
					              end if;
				  when EXECUTA => estado <= FETCH;
				  when others  => estado <=HALT;
			end case;
		end if;
	end process;  
	
	--
	-- GERACAO DAS MICROINSTRUCOES
	--
	process(estado)
	 begin
		
	  case estado is	  
		
		 -- gera a microinstrucao relativa ao ciclo de fetch
		 when FETCH =>  uins <= (c1, c0,c0, c1, c0,c0,c0,c0, c0,c0, c1,c0,c0,c0, c1,c1, incB );
					
		 when EXECUTA =>	 
			
			if  op1=x"0"   then
				uins <= (c0, c0,c1, c0, c0,c0,c0,c0, c1,c1, c1,c0,c0,c0, c0,c0, add );      -- soma 
			elsif  op1=x"1"   then
				uins <= (c0, c0,c1, c0, c0,c0,c0,c0, c1,c1, c1,c0,c0,c0, c0,c0, sub );      -- subtracao    
			elsif  op1=x"2"   then
				uins <= (c0, c0,c1, c0, c0,c0,c0,c0, c1,c1, c1,c0,c0,c0, c0,c0, e );        -- e 	   
			elsif  op1=x"3"   then
				uins <= (c0, c0,c1, c0, c0,c0,c0,c0, c1,c1, c1,c0,c0,c0, c0,c0, ou );       -- ou 	   
			elsif  op1=x"4"   then
				uins <= (c0, c0,c1, c0, c0,c0,c0,c0, c1,c1, c1,c0,c0,c0, c0,c0, oux );      -- ou exclusivo 	   
			elsif  op1=x"5"   then
				uins <= (c0, c1,c1, c0, c0,c0,c0,c0, c0,c1, c1,c0,c0,c0, c0,c0, addi );     -- soma imediato 	    	
			elsif  op1=x"6"   then
				uins <= (c0, c1,c1, c0, c0,c0,c0,c0, c0,c1, c1,c0,c0,c0, c0,c0, subi );     -- sub imediato 	   	   
			elsif  op1=x"7"   then
				uins <= (c0, c1,c1, c0, c0,c0,c0,c0, c0,c1, c1,c0,c0,c0, c0,c0, cte_L );    -- LDL 	   
			elsif  op1=x"8"   then
				uins <= (c0, c1,c1, c0, c0,c0,c0,c0, c0,c1, c1,c0,c0,c0, c0,c0, cte_H );    -- LDH	
			elsif  op1=x"A" and op3=x"1"  then	  
				uins <= (c0, c0,c1, c0, c1,c1,c1,c1, c1,c1, c0,c1,c0,c0, c1,c1, passaA );   -- pop desempilha registrador				
			elsif  op1=x"B" and op2=x"1"  then	  
				uins <= (c0, c0,c0, c0, c1,c0,c1,c0, c1,c1, c0,c1,c1,c0, c1,c0, passaA );   -- push empilha registrador				
			elsif  op1=x"A" and op3=x"2"  then	  
				uins <= (c0, c0,c1, c0, c0,c0,c0,c0, c1,c1, c0,c0,c0,c0, c1,c1, passaA );   -- LE DA RAN				
			elsif  op1=x"B" and op2=x"2"  then	  
				uins <= (c0, c0,c0, c0, c0,c0,c0,c0, c1,c1, c0,c0,c1,c0, c1,c0, passaA );   -- GRAVA NA RAM				
			elsif  op1=x"C" and op3=x"0"  then	  
				uins <= (c0, c0,c1, c0, c0,c0,c0,c0, c1,c1, c1,c0,c0,c0, c0,c0, sl );   -- sl
			elsif  op1=x"C" and op3=x"1"  then	  
				uins <= (c0, c0,c1, c0, c0,c0,c0,c0, c1,c1, c1,c0,c0,c0, c0,c0, sr );   -- sr
			elsif  op1=x"C" and op3=x"2"  then	  
				uins <= (c0, c0,c1, c0, c0,c0,c0,c0, c1,c1, c1,c0,c0,c0, c0,c0, notb );   -- not			
			elsif  op1=x"D" and (op2=x"0" or op2=x"1" or op2=x"2" or op2=x"3" )   then	   
				uins <= (c0, c0, c0, c0, c0,c0,c0,c0, c1,c1, c0,c0,c0,c0, c0,c0, passaA );  -- set os flags		
			elsif  op1=x"D" and op2=x"4"  then	  
				uins <= (c0, c0,c0, c0, c1,c0,c0,c0, c1,c1, c0,c0,c0,c0, c0,c0, passaA );    -- carrega SP			
			elsif  op1=x"D" and op2=x"6"  then	  
				uins <= (c0, c0,c0, c1, c1,c1,c1,c1, c0,c0, c0,c1,c0,c0, c1,c1, passaA );    -- RTS				
			elsif  op1=x"D" and op2=x"9"  and flag='1' then	  
				uins <= (c0, c0,c0, c1, c1,c0,c1,c0, c1,c0, c1,c1,c0,c1, c1,c0, add );    -- salta sub relativo		
			elsif  op1=x"D" and op2=x"B"  and flag='1' then	  
				uins <= (c0, c0,c0, c1, c1,c0,c1,c0, c1,c0, c1,c1,c0,c1, c1,c0, passaA );    -- salta para subrotina - reg
			elsif  op1=x"F" and flag='1' then	  
				uins <= (c0, c0,c0, c1, c1,c0,c1,c0, c0,c0, c1,c1,c0,c1, c1,c0, jmp );    -- salta para subrotina - imediato
			elsif  op1=x"D" and op2=x"8"  and flag='1' then	  
				uins <= (c0, c0,c0, c1, c0,c0,c0,c0, c1,c0, c1,c0,c0,c0, c0,c0, add );    -- salta relativo									   
			elsif  op1=x"D" and op2=x"A"  and flag='1' then	  
				uins <= (c0, c0,c0, c1, c0,c0,c0,c0, c1,c0, c1,c0,c0,c0, c0,c0, passaA );    -- salta  reg			
			elsif  op1=x"E" and flag='1' then	  
				uins <= (c0, c0,c0, c1, c0,c0,c0,c0, c0,c0, c1,c0,c0,c0, c0,c0, jmp );    -- salta imediato				
			else                                                                          -- NOP 
				uins <= (c0, c0, c0, c0, c0,c0,c0,c0, c0,c0, c0,c0,c0,c0, c0,c0, passaA );        				
			end if;
			
		 when others => uins <= (c0, c0, c0, c0, c0,c0,c0,c0, c0,c0, c0,c0,c0,c0, c0,c0, passaA );
	   end case;      
	end process;   
	
end controle;

--------------------------------------------------------------------------
-- Uniao dos componentes da r6
--------------------------------------------------------------------------
library IEEE;
use IEEE.Std_Logic_1164.all;
use work.r6.all;

entity processador is
	port(ck,rst: in std_logic;
		data: inout reg16;
		address: out reg16;
		ce,rw: out std_logic
		);
end processador;   

architecture processador of processador is
	
	component controle 
		port( uins:   out microinstrucao;
			rst,ck: in std_logic;
			flag:   in std_logic;
			ir:     in reg16
			);
	end component;   
	
	component datapath
		port( uins: in microinstrucao;
			ck, rst: in std_logic;
			instrucao, endereco : out reg16;
			data: inout reg16;
			flag: out std_logic
			); 
	end component;     
	
	signal flag: std_logic;
	signal uins: microinstrucao;
	signal ir: reg16;
	
	begin
	
	dp: datapath port map(uins=>uins,ck=>ck,rst=>rst,instrucao=>ir,endereco=>address,data=>data, flag=>flag);
	ctrl: controle port map(uins=>uins,rst=>rst,ck=>ck, flag=>flag,ir=>ir);
	
	ce <= uins.ce;        
	rw <= uins.rw;        
	
end processador;
