-------------------------------------------------------------------------
--  Fernando Moraes - ultima alteracao - 01/10/2000	
--  Ney Calazans - ultima alteracao - 16/03/2001	
-------------------------------------------------------------------------

-------------------------------------------------------------------------
--  TEST_BENCH PARA SIMULACAO DO PROCESSADOR R6
-------------------------------------------------------------------------
library ieee;
use ieee.STD_LOGIC_UNSIGNED.all;
use ieee.std_logic_1164.all;
use STD.TEXTIO.all;


entity r6_tb is
end r6_tb;

architecture TB_ARCHITECTURE of r6_tb is
	
	component processador
		port(ck,rst: in std_logic;
			data: inout std_logic_vector(15 downto 0);
			address: out std_logic_vector(15 downto 0);
			ce,rw: out std_logic
			);
	end component; 
	
	-- memoria de largura de 16 bits - enderecavel a palavra - MUITA atencao  !!
	type mem1 is array (0 to 511) of std_logic_vector(15 downto 0);
	constant TAMaddress : integer := 511;	 
	
	signal go, rw, ce, ck, rst : std_logic;
	signal data, ad, ins :  std_logic_vector(15 downto 0);	
	signal endereco : integer;
	
	signal RAM : mem1;
	
	file ARQ : TEXT open READ_MODE is "t_asm_R6.txt";
	begin
	
	UUT : processador
	port map
		(ck => ck, rst => rst, data=>data, address=>ad, ce=>ce, rw=>rw);
	
	--
	-- gera o reset
	--	  
	rst <='1', '0' after 5ns;	
	
	--
	-- gera o clock
	--	  
	process
		begin
		ck <= '1', '0' after 10ns;
		wait for 20ns;
	end process;
	
	-- le da memoria
	data <= RAM(CONV_INTEGER(ad)) when ad<(TAMaddress + 1) and ce='1' and rw='1' else (others=>'Z');
	
	-- escreve na memoria, de maneira sincrona, como nos registradores
	process(go, ck, ce, rw)
		begin
		-- borda de subida do go - escreve na memoria
		if go'event and go='1' then
			if endereco>=0 and endereco<=TAMaddress then
				RAM(endereco) <= ins;
			end if;
			-- borda de descida de clock
		elsif ck'event and ck='0' and ce='1' and rw='0' then	
			if CONV_INTEGER(ad)>=0 and CONV_INTEGER(ad)<=TAMaddress then
				RAM(CONV_INTEGER(ad)) <= data;	
			end if;
		end if;
	end process;
	
	-- realiza a carga na memoria quando acontece o reset
	process
		variable LINHA_ARQ : LINE;
		variable linha     : string(1 to 9);
		variable base, num : integer;
		variable bin       : STD_LOGIC_VECTOR(1 to 4);
		
		begin
		wait until rst = '1';
		
		while NOT (endfile(ARQ)) loop	-- end file checking
			readline(ARQ, LINHA_ARQ);	-- read line of a file
			read(LINHA_ARQ, linha);
			
			endereco <= 0;
			wait for 1ps;
			
			base := 1;
			for w in 4 downto 1 loop
				if w /= 4 then
					base := base * 16;
				end if;
				case linha(w) is
					when '0' => num := 0;
					when '1' => num := 1;
					when '2' => num := 2;
					when '3' => num := 3;
					when '4' => num := 4;
					when '5' => num := 5;
					when '6' => num := 6;
					when '7' => num := 7;
					when '8' => num := 8;
					when '9' => num := 9;
					when 'A' => num := 10;
					when 'B' => num := 11;
					when 'C' => num := 12;
					when 'D' => num := 13;
					when 'E' => num := 14;
					when 'F' => num := 15;
					when others => null;
				end case;
				
				endereco <= endereco + num * base;
				wait for 1ps;
			end loop;  -- end loop da conversao p/ decimal
			
			for w in 6 to 9 loop
				case linha(w) is  
					when '0' => bin := "0000";
					when '1' => bin := "0001";
					when '2' => bin := "0010";
					when '3' => bin := "0011";
					when '4' => bin := "0100";
					when '5' => bin := "0101";
					when '6' => bin := "0110";
					when '7' => bin := "0111";
					when '8' => bin := "1000";
					when '9' => bin := "1001";
					when 'A' => bin := "1010";
					when 'B' => bin := "1011";
					when 'C' => bin := "1100";
					when 'D' => bin := "1101";
					when 'E' => bin := "1110";
					when 'F' => bin := "1111";
					when others => null;
				end case;
				
				case w is
					when 6 => ins(15 downto 12) <= bin;
					when 7 => ins(11 downto 8)  <= bin;
					when 8 => ins(7 downto 4)   <= bin;
					when 9 => ins(3 downto 0)   <= bin;
				end case;
				
			end loop;  -- end loop da conversao p/ binario
			
			wait for 1ps;
			go <= '1';    -- este sinal de go habilita a escrita na memoria
			
			wait for 1ps;
			go <= '0';    -- tira o go
			
		end loop; -- end loop da leitura do arquivo
	end process;
	
end TB_ARCHITECTURE;